@@ -233,6 +233,26 @@ class TSimCNE:
         need to point to the beton file string and the size
         information cannot be inferred from that).
 
+    :param 1 devices: The number of devices/accelerators to use (with
+        the PL Trainer).  Will be passed on as is.  Consider this
+        parameter experimental, the effects of using multiple GPUs is
+        not entirely clear (but it should probably be safe to do so).
+
+        Currently, the learning rate is not adjusted to account for
+        multiple devices, please do so yourself; see the `PL
+        documentation
+        <https://lightning.ai/docs/pytorch/stable/accelerators/gpu_faq.html>`__
+        about it.
+
+    :param dict | None trainer_kwargs: The keyword arguments to pass
+        to the Trainer, to use during training.  By default the keys
+        ``gradient_clip_val=4`` and
+        ``gradient_clip_algorithm="value"`` will be set, but can be
+        overridden by passing in a custom dict.  The values will be
+        set regardless of whether you pass in a ``dict`` or not, so if
+        you want to disable gradient clipping you need to override the
+        values.
+
     :param int=8 num_workers: The number of workers for creating the
         dataloader.  Will be passed to the pytorch DataLoader
         constructor.
@@ -268,6 +288,8 @@ class TSimCNE:
         warmup="auto",
         freeze_schedule="only_linear",
         image_size=None,
+        devices=1,
+        trainer_kwargs=None,
         num_workers=8,
         use_ffcv="auto",
         float32_matmul_precision="medium",
@@ -288,6 +310,8 @@ class TSimCNE:
         self.warmup = warmup
         self.freeze_schedule = freeze_schedule
         self.image_size = image_size
+        self.devices = devices
+        self.trainer_kwargs = trainer_kwargs
         self.num_workers = num_workers
         self.use_ffcv = use_ffcv
         self.float32_matmul_precision = float32_matmul_precision
@@ -349,6 +373,24 @@ class TSimCNE:
                 f"but got {self.freeze_schedule}."
             )
 
+        if self.devices != 1 and self.lr == "auto_batch":
+            import warnings
+
+            warnings.warn(
+                "devices is not 1, but the learning rate has not been adjusted."
+                "  Please see https://"
+                "lightning.ai/docs/pytorch/stable/accelerators/gpu_faq.html "
+                "for how to set the learning rate when using multiple devices"
+            )
+
+        trainer_kwargs = dict(
+            gradient_clip_val=4, gradient_clip_algorithm="value"
+        )
+        if self.trainer_kwargs is None:
+            self.trainer_kwargs = trainer_kwargs
+        else:
+            self.trainer_kwargs = trainer_kwargs.update(self.trainer_kwargs)
+
     @staticmethod
     def check_ffcv(use_ffcv):
         if use_ffcv:
@@ -438,6 +480,8 @@ class TSimCNE:
                 lr_scheduler_name=self.lr_scheduler,
                 lr=lr,
                 warmup=warmup_epochs,
+                pretrain_out_dim=self.pretrain_out_dim,
+                out_dim=self.out_dim,
                 use_ffcv=self.use_ffcv,
             )
             if n_stage == 0:
@@ -448,7 +492,6 @@ class TSimCNE:
                     metric=self.metric,
                     backbone=self.backbone,
                     projection_head=self.projection_head,
-                    out_dim=self.pretrain_out_dim,
                     **train_kwargs,
                 )
 
@@ -465,7 +508,6 @@ class TSimCNE:
                     model=model,
                     loss=p.loss,
                     metric=p.metric,
-                    out_dim=self.out_dim,
                     **train_kwargs,
                 )
 
@@ -475,10 +517,13 @@ class TSimCNE:
                     model=model,
                     loss=p.loss,
                     metric=p.metric,
-                    out_dim=self.out_dim,
                     **train_kwargs,
                 )
-            trainer = pl.Trainer(max_epochs=n_epochs, devices=1)
+            trainer = pl.Trainer(
+                max_epochs=n_epochs,
+                devices=self.devices,
+                **self.trainer_kwargs,
+            )
             trainer.fit(model=plmodel, train_dataloaders=train_dl)
             self.models.append(plmodel)
             self.trainers.append(trainer)
@@ -525,6 +570,8 @@ class TSimCNE:
         elif not return_labels and return_backbone_feat:
             return Y, backbone_features
         elif return_labels and not return_backbone_feat:
+            # XXX: this for some reason changes the labels; but I
+            # don't know what causes this!
             labels = torch.hstack([lbl for _, lbl in loader])
             return Y, labels
         else:
@@ -578,6 +625,7 @@ class TSimCNE:
             )
         else:
             import ffcv
+            from ffcv.fields.basics import IntDecoder
 
             if train_or_test:
                 self.data_transform2 = get_transforms_unnormalized(
@@ -585,7 +633,7 @@ class TSimCNE:
                 )
 
                 self.label_pipeline = [
-                    ffcv.fields.basics.IntDecoder(),
+                    IntDecoder(),
                     ffcv.transforms.ToTensor(),
                     ffcv.transforms.Squeeze(),
                 ]
